from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import UserPassesTestMixin
from django.db.models import Q
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic import DetailView

from cruditor.views import (
    CruditorAddView,
    CruditorChangeView,
    CruditorDeleteView,
    CruditorListView,
)
from django_filters.views import FilterView
from django_tables2.views import SingleTableMixin

from benevalibre.base.tables import (
    AssociatedEngagedSelfBenevaloTable,
    EngagementIndexTable,
)
from benevalibre.utils.html import fa_icon
from benevalibre.utils.views import (
    CruditorPageMixin,
    EmailNotifySuccessMixin,
    PageMixin,
)

from . import filters, forms, models, tables


def get_management_buttons(association, user):
    """
    Return the management actions that `user` has on `association` as a list
    of buttons - e.g. to be displayed in the title buttons.
    """
    buttons = []

    def _reverse(viewname):
        return reverse(viewname, args=[association.id])

    if association.can_manage_engagements(user):
        buttons.append(
            {
                'label': fa_icon('users', "Gérer les bénévoles"),
                'url': _reverse('association:engagement:index'),
            }
        )
    elif association.can_list_users(user):
        buttons.append(
            {
                'label': fa_icon('users', "Voir les bénévoles"),
                'url': _reverse('base:association:user'),
            }
        )
    if association.can_manage_benevalos(user):
        buttons.append(
            {
                'label': fa_icon('gift', "Gérer les bénévolats"),
                'url': _reverse('base:association:benevalo:index'),
            }
        )
    if association.can_manage_levels(user):
        buttons.append(
            {
                'label': fa_icon(
                    'tachometer', "Gérer les niveaux de bénévolat"
                ),
                'url': _reverse('association:level:index'),
            }
        )
    if association.can_manage_projects(user):
        buttons.append(
            {
                'label': fa_icon(
                    'wrench', "Gérer les projets de l'association"
                ),
                'url': _reverse('association:project:index'),
            }
        )
    if association.can_manage_categories(user):
        buttons.append(
            {
                'label': fa_icon('tags', "Gérer les catégories de bénévolat"),
                'url': _reverse('association:category:index'),
            }
        )
    if association.can_manage_roles(user):
        buttons.append(
            {
                'label': fa_icon(
                    'key-modern', "Gérer les roles de l'association"
                ),
                'url': _reverse('association:role:index'),
            }
        )
    if association.can_manage_association(user):
        buttons.append(
            {
                'label': fa_icon('edit', "Modifier l'association"),
                'url': _reverse('association:update'),
            }
        )

    return buttons


class AssociationRelatedMixin:
    """
    Mixin to apply to the views related to an existing `Association`. The
    object with the id given by the `association` keyword argument will be
    fetched at setup and stored in the `association` property.
    """

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.association = self.get_association()

    def get_association(self):
        return get_object_or_404(
            models.Association.objects.viewable_by(
                self.request.user
            ).distinct(),
            pk=self.kwargs['association'],
        )

    def get_title(self):
        return "{} pour « {} »".format(super().get_title(), self.association)


class AssociationRelatedFormMixin:
    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        form.instance.association = self.association
        return form


# ASSOCIATIONS
# ------------------------------------------------------------------------------


class AssociationIndex(PageMixin, FilterView):
    """
    La page « Liste des associations » vous permet de voir les associations
    enregistrées et visibles publiquement au sein de l'application. En cliquant
    sur l'une d'entre elle, vous accéderez à ses détails.

    Il est également possible à partir de cette page d’:doc:`enregistrer une
    association <enregistrer_association>`.

    Vous pouvez accéder à cette page depuis le menu principal ou depuis la
    page d'accueil si vous n'êtes pas connecté⋅e.
    """

    template_name = 'association/association_list.html'
    title = "Liste des associations"
    filterset_class = filters.AssociationIndexFilter

    def get_queryset(self):
        return (
            models.Association.objects.viewable_by(self.request.user)
            .distinct()
            .order_by('-is_active', 'name')
        )

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        # FIXME: centraliser les permissions de l'utilisateur et récupérer
        # plus finement celle d'ajouter une association en tenant compte de
        # la modération en plus
        context['can_add_association'] = settings.OPEN_REGISTRATION
        return context


class AssociationCreate(
    EmailNotifySuccessMixin, CruditorPageMixin, CruditorAddView
):
    """
    Créer une association nécessite de remplir le formulaire idoine.

    Si une association existe déjà sous le même nom, un avertissement sera
    levé. Il est néanmoins possible de forcer l'avertissement en re-validant
    le formulaire.

    Si l'instance de **Bénévalibre** modère les inscriptions d'association :

    * les administrateurs de la plateforme seront prévenus pour donner suite ;
    * vous serez redirigé vers l'accueil.

    Autrement :

    * l'association est créée immédiatement ;
    * vous êtes redirigés vers la page de présentation de l'association.

    Dans tous les cas :

    * les rôles par défaut sont créés et attachés à l'association ;
    * les catégories par défaut sont créées et attachées à l'association ;
    * les niveaux par défaut sont créées et attachées à l'association ;
    * l'utilisateur à l'origine de l'enregistrement de l'association devient
      dirigeant de l'association ;
    """

    model = models.Association
    form_class = forms.CreateForm
    title = "Ajouter une association"

    notify_users_query = Q(is_superuser=True)
    notify_template = 'association/registration_moderation_email.txt'
    notify_subject_template = 'association/registration_moderation_subject.txt'

    def get_notify_condition(self):
        # Nota Bene: could be static attribute but not during tests
        return settings.MODERATED_REGISTRATION

    def dispatch(self, request, *args, **kwargs):
        if not settings.OPEN_REGISTRATION:
            raise Http404
        return super().dispatch(request, *args, **kwargs)

    def check_multiple(self, form):
        warned = self.request.session.get('warned', [])
        if (
            self.object.name not in warned
            and self.model.objects.filter(name=self.object.name).exists()
        ):
            messages.warning(
                self.request,
                "Une association existe déjà sous ce nom. Êtes vous certain "
                "de vouloir poursuivre la création ?",
            )
            warned.append(self.object.name)
            self.request.session['warned'] = warned
            # Re-render the page squeezing CruditorMixin's form_invalid.
            return self.render_to_response(
                self.get_context_data(form=form, formset_errors=False)
            )
        elif self.object.name in warned:
            # User has been warned, he force.
            warned.remove(self.object.name)
            self.request.session['warned'] = warned

    def form_valid(self, form):
        self.object = form.save(commit=False)
        resend_form = self.check_multiple(form)
        if resend_form:
            return resend_form
        self.object = form.save()
        if settings.MODERATED_REGISTRATION:
            self.success_message = (
                "Votre demande a bien été reçue et soumise à modération. "
                "Les modérat⋅eurs⋅rices de l'instance donneront suite "
                "dès que possible."
            )
        # We setup the creator with the local administrator role
        role = self.object.role_set.get_dirigeant()
        self.object.engagement_set.create(user=self.request.user, role=role)
        return super().form_valid(form)

    def get_success_url(self):
        if settings.MODERATED_REGISTRATION:
            return reverse('base:home')
        else:
            return reverse('association:detail', args=[self.object.pk])


class AssociationDetail(PageMixin, SingleTableMixin, DetailView):
    template_name = 'association/association_detail.html'
    context_table_name = 'benevalo_table'
    table_class = AssociatedEngagedSelfBenevaloTable

    def get_queryset(self):
        return models.Association.objects.viewable_by(
            self.request.user
        ).distinct()

    def get_title(self):
        return self.object.name

    def get_titlebuttons(self):
        management_buttons = get_management_buttons(
            self.object, self.request.user
        )
        if management_buttons:
            return [
                {
                    'label': fa_icon('gear', "Gérer…"),
                    'submenu': management_buttons,
                }
            ]
        return []

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['can_engage'] = self.object.can_self_engage(self.request.user)
        return context

    def get_table(self, **kwargs):
        if not self.object.can_self_create_benevalo(self.request.user):
            return None
        return super().get_table(**kwargs)

    def get_table_data(self):
        return self.object.benevalo_set.by_user(self.request.user)


class AssociationUpdate(CruditorPageMixin, CruditorChangeView):
    model = models.Association
    form_class = forms.UpdateForm
    title = "Modifier une association"

    def get_object(self, **kwargs):
        asso = super().get_object(**kwargs)
        if not asso.can_manage_association(self.request.user):
            raise Http404
        return asso

    def get_success_url(self):
        return reverse('association:detail', args=[self.object.pk])

    def get_delete_url(self):
        return reverse('association:delete', args=[self.object.pk])

    def form_valid(self, form):
        self.object = form.save()
        if not self.object.is_active and self.request.user.is_superuser:
            self.object.is_active = True
            # FIXME: notifier le créateur (en option)
        return super().form_valid(form)


class AssociationDelete(CruditorPageMixin, CruditorDeleteView):
    model = models.Association
    title = "Supprimer une association"

    def get_object(self, **kwargs):
        asso = super().get_object(**kwargs)
        if not asso.can_manage_association(self.request.user):
            raise Http404
        return asso

    def get_success_url(self):
        return reverse('base:home')


# ASSOCIATION'S PROJECTS
# ------------------------------------------------------------------------------


class ProjectViewMixin(
    AssociationRelatedMixin, UserPassesTestMixin, CruditorPageMixin
):
    model = models.Project

    def test_func(self):
        return self.association.can_manage_projects(self.request.user)

    def get_queryset(self):
        return self.association.project_set.all()

    def get_success_url(self):
        return reverse('association:project:index', args=[self.association.id])


class ProjectIndex(ProjectViewMixin, CruditorListView):
    table_class = tables.ProjectIndexTable
    title = "Liste des projets"
    create_titlebutton_label = "Ajouter un projet"

    def get_create_url(self):
        return reverse(
            'association:project:create', args=[self.association.id]
        )


class ProjectCreate(
    ProjectViewMixin, AssociationRelatedFormMixin, CruditorAddView
):
    form_class = forms.ProjectCreateForm
    title = "Ajouter un projet"


class ProjectUpdate(
    ProjectViewMixin, AssociationRelatedFormMixin, CruditorChangeView
):
    form_class = forms.ProjectUpdateForm

    def get_delete_url(self):
        return reverse(
            'association:project:delete',
            args=[self.association.id, self.object.pk],
        )


class ProjectDelete(ProjectViewMixin, CruditorDeleteView):
    pass


# ASSOCIATION'S ROLES
# ------------------------------------------------------------------------------


class RoleViewMixin(
    AssociationRelatedMixin, UserPassesTestMixin, CruditorPageMixin
):
    model = models.Role

    def test_func(self):
        return self.association.can_manage_roles(self.request.user)

    def get_queryset(self):
        return self.association.role_set.all()

    def get_success_url(self):
        # We may have lost our capability
        if not self.test_func():
            return reverse('base:board')
        return reverse('association:role:index', args=[self.association.id])


class RoleIndex(RoleViewMixin, CruditorListView):
    table_class = tables.RoleIndexTable
    title = "Liste des rôles des bénévoles"
    create_titlebutton_label = "Ajouter un rôle"

    def get_create_url(self):
        return reverse('association:role:create', args=[self.association.id])


class RoleCreate(RoleViewMixin, AssociationRelatedFormMixin, CruditorAddView):
    form_class = forms.RoleCreateForm
    title = "Ajouter un rôle"


class RoleUpdate(
    RoleViewMixin, AssociationRelatedFormMixin, CruditorChangeView
):
    form_class = forms.RoleUpdateForm

    def get_delete_url(self):
        return reverse(
            'association:role:delete',
            args=(self.association.id, self.object.pk),
        )


class RoleDelete(RoleViewMixin, CruditorDeleteView):
    pass


# ASSOCIATION'S LEVELS
# ------------------------------------------------------------------------------


class LevelViewMixin(
    AssociationRelatedMixin, UserPassesTestMixin, CruditorPageMixin
):
    model = models.Level

    def test_func(self):
        return self.association.can_manage_levels(self.request.user)

    def get_queryset(self):
        return self.association.level_set.all()

    def get_success_url(self):
        return reverse('association:level:index', args=[self.association.id])


class LevelIndex(LevelViewMixin, CruditorListView):
    table_class = tables.LevelIndexTable
    title = "Liste des niveaux de bénévolat"
    create_titlebutton_label = "Ajouter un niveau de bénévolat"

    def get_create_url(self):
        return reverse('association:level:create', args=[self.association.id])


class LevelCreate(
    LevelViewMixin, AssociationRelatedFormMixin, CruditorAddView
):
    form_class = forms.LevelCreateForm
    title = "Ajouter un niveau de bénévolat"


class LevelUpdate(
    LevelViewMixin, AssociationRelatedFormMixin, CruditorChangeView
):
    form_class = forms.LevelUpdateForm

    def get_delete_url(self):
        return reverse(
            'association:level:delete',
            args=(self.association.id, self.object.pk),
        )


class LevelDelete(LevelViewMixin, CruditorDeleteView):
    pass


# ASSOCIATION'S CATEGORIES
# ------------------------------------------------------------------------------


class CategoryViewMixin(
    AssociationRelatedMixin, UserPassesTestMixin, CruditorPageMixin
):
    model = models.Category

    def test_func(self):
        return self.association.can_manage_categories(self.request.user)

    def get_queryset(self):
        return self.association.category_set.all()

    def get_success_url(self):
        return reverse(
            'association:category:index', args=[self.association.id]
        )


class CategoryIndex(CategoryViewMixin, CruditorListView):
    table_class = tables.CategoryIndexTable
    title = "Liste des catégories de bénévolat"
    create_titlebutton_label = "Ajouter une catégorie de bénévolat"

    def get_create_url(self):
        return reverse(
            'association:category:create', args=[self.association.id]
        )


class CategoryCreate(
    CategoryViewMixin, AssociationRelatedFormMixin, CruditorAddView
):
    form_class = forms.CategoryCreateForm
    title = "Ajouter une catégorie de bénévolat"


class CategoryUpdate(
    CategoryViewMixin, AssociationRelatedFormMixin, CruditorChangeView
):
    form_class = forms.CategoryUpdateForm

    def get_delete_url(self):
        return reverse(
            'association:category:delete',
            args=(self.association.id, self.object.pk),
        )


class CategoryDelete(CategoryViewMixin, CruditorDeleteView):
    pass


# ASSOCIATION'S ENGAGEMENTS
# ------------------------------------------------------------------------------


class EngagementViewMixin(
    AssociationRelatedMixin, UserPassesTestMixin, CruditorPageMixin
):
    model = models.Engagement

    def test_func(self):
        return self.association.can_manage_engagements(self.request.user)

    def get_queryset(self):
        return self.association.engagement_set.all()

    def get_success_url(self):
        # We may have lost our capability
        if not self.test_func():
            return reverse('base:board')
        return reverse(
            'association:engagement:index', args=[self.association.id]
        )


class EngagementIndex(EngagementViewMixin, CruditorListView):
    table_class = EngagementIndexTable
    title = "Liste des bénévoles"


class EngagementUpdate(
    EmailNotifySuccessMixin,
    EngagementViewMixin,
    AssociationRelatedFormMixin,
    CruditorChangeView,
):
    form_class = forms.EngagementUpdateForm

    notify_template = 'association/engagement_update_email.txt'
    notify_subject_template = 'association/engagement_update_subject.txt'

    def get_notify_condition(self):
        return self.object.user != self.request.user

    def get_notify_users_query(self):
        return Q(id=self.object.user.id)

    def get_delete_url(self):
        return reverse(
            'association:engagement:delete',
            args=(self.association.id, self.object.pk),
        )

    def form_valid(self, form):
        self.object = form.save()
        if not self.object.is_active:
            self.object.is_active = True
        return super().form_valid(form)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        user_role = self.association.get_role(self.request.user)
        if self.object.is_active:
            managed_user_role = self.association.get_role(self.object.user)
        else:
            managed_user_role = self.association.role_set.get_benevole()
        deleguables = self.association.role_set.delegables_by(user_role)

        kwargs['role_set'] = (
            deleguables
            if managed_user_role in deleguables
            else self.association.role_set.filter(pk=managed_user_role.pk)
        )
        return kwargs


class EngagementDelete(EngagementViewMixin, CruditorDeleteView):
    pass
