from datetime import timedelta

from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.duration import _get_duration_components
from django.utils.html import format_html
from django.utils.text import Truncator

import django_tables2 as tables
from django_tables2.utils import A

from benevalibre.association import models
from benevalibre.association.tables import AssociationRelativeActionsMixin
from benevalibre.benevalo.models import Benevalo
from benevalibre.utils.html import fa_icon

from . import forms


class ModerationBadgedLinkColumn(tables.Column):
    def __init__(
        self, urlname, attrs={}, text=None, args=None, enabler=None, **kwargs
    ):
        self.text = text
        self.urlname = urlname
        self.args = args
        self.attrs = attrs
        self.enabler = enabler
        super().__init__(**kwargs)

    def render(self, value, *args, **kwargs):
        record = kwargs['record']
        attrs = self.attrs.get('a', {})
        cls = attrs.get('class', '')
        args = [a.resolve(record) for a in self.args]
        url = reverse(self.urlname, args=args)
        badge = ""
        if not self.enabler.resolve(record):
            badge = format_html(
                "<span class='ml-2 badge badge-pill badge-primary'>{}"
                "<span class='sr-only'>(en modération)</span></span>",
                fa_icon('unlock-alt', margin=''),
            )

        return format_html(
            "<a class='{}' href='{}'>{}</a>{}", cls, url, value, badge
        )

    def value(self, value):
        return value


class AssociationIndexTable(tables.Table):
    name = ModerationBadgedLinkColumn(
        'association:detail', args=(A('pk'),), enabler=A('is_active')
    )
    logo = tables.TemplateColumn(
        template_name='base/tables/association_logo.html',
        verbose_name='',
        orderable=False,
        accessor=A('logo_thumb'),
    )
    engagement = tables.Column(
        accessor=A('pk'), verbose_name='Devenir bénévole', orderable=False
    )

    def render_engagement(self, **kwargs):
        return render_to_string(
            'base/includes/engagement_form.html',
            {
                'csrf_token': self.context['csrf_token'],
                'record': kwargs['record'],
                'form': forms.AssociationEngageForm(),
                'active': kwargs['record'].can_self_engage(
                    self.context['user']
                ),
                'engaged': kwargs['record'].is_engaged(self.context['user']),
                'user': self.context['user'],
            },
        )

    def render_description(self, value):
        return Truncator(value).words(8, truncate=' …')

    class Meta:
        model = models.Association
        fields = ('logo', 'name', 'description', 'engagement')


class AssociationModerationTable(AssociationIndexTable):
    engagement = None
    inspect = tables.LinkColumn(
        'association:update',
        verbose_name="Donner suite",
        text=fa_icon('eye', "Inspecter"),
        args=(A('pk'),),
        orderable=False,
        attrs={'a': {'class': 'btn btn-primary update-link'}},
    )
    user = tables.Column(accessor=A('engagement_set.all.0.user'))
    avatar = tables.TemplateColumn(
        accessor=A('engagement_set.all.0.user'),
        template_name='base/tables/user_avatar.html',
        verbose_name='',
        orderable=False,
    )

    class Meta:
        model = models.Association
        fields = ('logo', 'name', 'user', 'avatar', 'description', 'inspect')


class EngagementModerationTable(tables.Table):
    logo = tables.TemplateColumn(
        template_name='base/tables/association_logo.html',
        verbose_name='',
        orderable=False,
        accessor=A('association.logo_thumb'),
    )
    association = tables.LinkColumn(
        'association:detail',
        verbose_name="Association",
        args=(A('association.pk'),),
        attrs={'a': {'class': 'association-link'}},
    )

    inspect = tables.LinkColumn(
        'association:engagement:update',
        verbose_name="Donner suite",
        text=fa_icon('eye', "Inspecter"),
        args=(A('association.pk'), A('pk')),
        orderable=False,
        attrs={'a': {'class': 'btn btn-primary update-link'}},
    )

    avatar = tables.TemplateColumn(
        accessor=A('user'),
        template_name='base/tables/user_avatar.html',
        verbose_name='',
        orderable=False,
    )

    class Meta:
        model = models.Engagement
        fields = ('logo', 'association', 'avatar', 'user', 'role', 'inspect')


def pluralize(text, count):
    return text if count < 2 else "{}s".format(text)


def footer_count_association(table):
    count = len(set([o.association for o in table.data]))
    return pluralize("{} association".format(count), count)


class EnabledLinkColumn(tables.Column):
    def __init__(
        self, urlname, attrs={}, text=None, args=None, enablers=None, **kwargs
    ):
        self.text = text
        self.urlname = urlname
        self.args = args
        self.attrs = attrs
        self.enablers = enablers
        super().__init__(accessor=enablers[0], **kwargs)

    def render(self, value, *args, **kwargs):
        record = kwargs['record']
        text = self.text
        attrs = self.attrs.get('a', {})
        cls = attrs.get('class', '')
        if all([e.resolve(record) for e in self.enablers]):
            args = [a.resolve(record) for a in self.args]
            url = reverse(self.urlname, args=args)
        else:
            url = '#'
            cls += ' disabled'

        return format_html("<a class='{}' href='{}'>{}</a>", cls, url, text)


class ModeratedTextColumn(tables.Column):
    def render(self, value, **kwargs):
        record = kwargs['record']
        if not record.is_active or not record.association.is_active:
            return "En modération"
        return super().render(value)


class MyEngagementTable(tables.Table):
    logo = tables.TemplateColumn(
        template_name='base/tables/association_logo.html',
        verbose_name='',
        orderable=False,
        accessor=A('association.logo_thumb'),
    )
    association = ModerationBadgedLinkColumn(
        'association:detail',
        verbose_name="Association",
        args=(A('association.pk'),),
        attrs={'a': {'class': 'association-link'}},
        footer=footer_count_association,
        enabler=A('association.is_active'),
    )
    create_benevalo = EnabledLinkColumn(
        'base:association:self-benevalo:create',
        verbose_name="Saisir un bénévolat",
        text=fa_icon('plus-circle', text="Ajouter"),
        args=(A('association.id'),),
        attrs={'a': {'class': 'btn btn-primary create_benevalo-link'}},
        orderable=False,
        enablers=[A('is_active'), A('association.is_active')],
    )
    benevolats = EnabledLinkColumn(
        'base:association:self-benevalo:index',
        verbose_name="Liste des bénévolats",
        text=fa_icon('list', text="Lister"),
        args=(A('association.id'),),
        orderable=False,
        attrs={'a': {'class': 'btn btn-secondary benevolats-link'}},
        enablers=[A('is_active'), A('association.is_active')],
    )
    role = ModeratedTextColumn()

    class Meta:
        model = models.Engagement
        fields = (
            'logo',
            'association',
            'role',
            'create_benevalo',
            'benevolats',
        )


def format_decimal(d):
    return "{:d},{:02d}".format(int(d), int((d * 100) % 100))


def footer_distance(table):
    distance = sum(x.distance for x in table.data if x.distance)
    return "Total : {} km".format(format_decimal(distance))


def format_duration(delta, no_days=False):
    days, hours, minutes, _, _ = _get_duration_components(delta)
    if no_days:
        hours += 24 * days
        days = 0
    string = ''
    if minutes:
        string = '{} {}'.format(minutes, pluralize('minute', minutes))
    if hours:
        if string:
            string = ' et ' + string
        string = '{} {}'.format(hours, pluralize('heure', hours)) + string
    if days:
        if string:
            if ' et ' in string:
                string = ', ' + string
            else:
                string = ' et ' + string
        string = '{} {}'.format(days, pluralize('jour', days)) + string
    return string


def footer_duration(table):
    duration = sum(
        [x.duration for x in table.data if x.duration], timedelta(0)
    )
    return "Total : {}".format(format_duration(duration, no_days=True))


def footer_date(table):
    if table.data:
        date_min = min(
            [x.execution_date for x in table.data if x.execution_date]
        )
        date_max = max(
            [x.execution_date for x in table.data if x.execution_date]
        )
        if date_min != date_max:
            return "sur {}".format(format_duration(date_max - date_min))
        else:
            return ""


def footer_count(table):
    count = len(table.data)
    return pluralize("{} élément".format(count), count)


class AssociatedEngagedSelfBenevaloTable(tables.Table):
    title = ModerationBadgedLinkColumn(
        'base:association:self-benevalo:update',
        args=(A('association.pk'), A('pk')),
        footer=footer_count,
        enabler=A('is_active'),
    )
    execution_date = tables.Column(footer=footer_date)
    distance = tables.Column(footer=footer_distance)
    duration = tables.Column(footer=footer_duration)

    def render_distance(self, value):
        return format_decimal(value)

    def render_duration(self, value):
        return format_duration(value, no_days=True)

    class Meta:
        model = Benevalo
        fields = ('title', 'execution_date', 'duration', 'distance')


class BenevaloModerationTable(tables.Table):
    inspect = tables.LinkColumn(
        'base:association:benevalo:update',
        verbose_name="Donner suite",
        text=fa_icon('eye', "Inspecter"),
        args=(A('association.pk'), A('pk')),
        attrs={'a': {'class': 'btn btn-primary update-link'}},
        orderable=False,
    )
    avatar = tables.TemplateColumn(
        accessor=A('user'),
        template_name='base/tables/user_avatar.html',
        verbose_name='',
        orderable=False,
    )

    distance = tables.Column()
    duration = tables.Column()

    def render_distance(self, value):
        return format_decimal(value)

    def render_duration(self, value):
        return format_duration(value, no_days=True)

    class Meta:
        model = Benevalo
        fields = (
            'avatar',
            'user',
            'title',
            'execution_date',
            'duration',
            'distance',
            'inspect',
        )


class BenevaloExportTable(AssociatedEngagedSelfBenevaloTable):
    execution_end_date = tables.Column()
    en_moderation = tables.Column(
        verbose_name="En attente de modération", accessor=A('is_active')
    )
    heures = tables.Column(
        verbose_name="Heures, en décimal", accessor=A('duration')
    )

    def render_en_moderation(self, value):
        return not bool(value)

    def render_heures(self, value):
        hours = 24 * value.days + value.seconds / 3600
        return format_decimal(hours)

    class Meta:
        model = Benevalo
        fields = (
            'user',
            'association',
            'title',
            'description',
            'execution_date',
            'execution_end_date',
            'duration',
            'heures',
            'en_moderation',
            'distance',
            'category',
            'project',
            'level',
        )


class AssociatedEngagedBenevaloTable(AssociatedEngagedSelfBenevaloTable):
    actions_update_url = 'base:association:benevalo:update'
    actions_delete_url = 'base:association:benevalo:delete'

    title = ModerationBadgedLinkColumn(
        'base:association:benevalo:update',
        args=(A('association.pk'), A('pk')),
        footer=footer_count,
        enabler=A('is_active'),
    )


def footer_count_people(table):
    count = len(set([o.user for o in table.data]))
    return pluralize("{} bénévole".format(count), count)


class AssociatedBenevaloTable(AssociatedEngagedBenevaloTable):
    actions_update_url = 'base:association:benevalo:update'
    actions_delete_url = 'base:association:benevalo:delete'

    avatar = tables.TemplateColumn(
        accessor=A('user'),
        template_name='base/tables/user_avatar.html',
        verbose_name='',
        orderable=False,
    )
    user = tables.Column(footer=footer_count_people)

    class Meta:
        model = Benevalo
        fields = (
            'avatar',
            'user',
            'title',
            'execution_date',
            'duration',
            'distance',
            'actions',
        )


class EngagementIndexTable(AssociationRelativeActionsMixin, MyEngagementTable):
    actions_update_url = 'association:engagement:update'
    actions_delete_url = 'association:engagement:delete'

    association = None
    logo = None
    create_benevalo = None

    avatar = tables.TemplateColumn(
        accessor=A('user'),
        template_name='base/tables/user_avatar.html',
        verbose_name='',
        orderable=False,
    )
    benevolats = EnabledLinkColumn(
        'base:association:engagement:benevalo',
        verbose_name="Liste des bénévolats",
        text=fa_icon('list', text="Lister"),
        args=(A('association.id'), A('pk')),
        orderable=False,
        attrs={'a': {'class': 'btn btn-secondary benevolats-link'}},
        enablers=[A('is_active'), A('association.is_active')],
    )
    user = tables.Column(footer=footer_count_people)

    class Meta:
        model = models.Engagement
        fields = ('avatar', 'user', 'role', 'benevolats', 'actions')


class EngagementReadOnlyIndexTable(EngagementIndexTable):
    benevolats = None
    actions = None
    # FIXME : avatar

    class Meta:
        model = models.Engagement
        fields = ('user', 'role')
