from datetime import date, timedelta

from django.contrib.messages import get_messages
from django.urls import reverse, reverse_lazy

import pytest
from bs4 import BeautifulSoup as bs

from benevalibre.association.models import Association
from benevalibre.utils.tests import (
    RE_DELETE_BTN,
    count_text_in_content,
    is_login_view,
)


class TestHome:
    url = reverse_lazy('base:home')

    def test_anonymous(self, client):
        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, reverse('association:index'))

    @pytest.mark.django_db
    def test_user_redirect(self, client, user):
        client.force_login(user)

        response = client.get(self.url)
        assert response.status_code == 302
        assert response.url == reverse('base:board')


@pytest.mark.django_db
class TestAssociationEngage:
    board_url = reverse_lazy('base:board')

    @pytest.fixture(autouse=True)
    def setup_association(self, association):
        self.association = association
        self.url = reverse('base:association:engage', args=[association.id])

    def test_get_redirect(self, client, user):
        client.force_login(user)
        response = client.get(self.url, follow=True)
        assert response.resolver_match.view_name == 'association:index'
        assert not self.association.get_engagement(user)

    def test_engage(self, client, user):
        assert self.association.can_self_engage(user)

        client.force_login(user)
        response = client.post(self.url, {}, follow=True)
        assert response.status_code == 200
        assert self.association.get_engagement(user)
        assert count_text_in_content(response, "Mes engagements associatifs")

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "modération" not in messages[0]
        assert "acceptée" in messages[0]

    def test_association_engagement_twice(self, client, user):
        client.force_login(user)
        response = client.post(self.url, {}, follow=True)
        response = client.post(self.url, {}, follow=True)
        assert response.status_code == 200
        assert self.association.get_engagement(user)
        assert response.resolver_match.view_name == 'association:index'

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "déjà engagé dans cette association" in messages[0]

    def test_moderated_association_engagement(
        self, client, mailoutbox, user, engagement, foreign
    ):
        self.association.moderate_engagement = True
        self.association.save()
        engagement.role = self.association.role_set.get_dirigeant()
        engagement.save()

        assert not self.association.can_self_engage(user)

        # 1. on fait la demande d'engagement
        client.force_login(foreign)
        response = client.post(self.url, {}, follow=True)
        assert response.status_code == 200

        # 2. elle est suspendue en modération
        engagements_qs = foreign.engagement_set.filter(
            association=self.association
        )
        assert len(engagements_qs) == 1
        assert not engagements_qs.active().exists()
        assert count_text_in_content(response, "Mes engagements associatifs")
        assert count_text_in_content(response, "Enregistrer une association")

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "modération" in messages[0]

        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert len(table_row) == 1
        assert "AssoTest" in str(table_row[0])
        assert 'href="#"' in str(table_row[0])
        assert "disabled" in str(table_row[0])
        assert "En modération" in str(table_row[0])

        # 3. elle est visible à un modérat⋅eur⋅rice
        client.logout()
        client.force_login(user)
        response = client.get(self.board_url)
        assert count_text_in_content(response, "Modérations de bénévoles")

        # Qui reçoit une notification
        assert len(mailoutbox) == 1
        assert user.email in mailoutbox[0].to
        assert user.get_short_name() in mailoutbox[0].body
        assert "Validation d'un nouvel engagement" in mailoutbox[0].subject

        # 4. on donne suite
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a', class_='update-link')[-1]
            .attrs['href']
        )
        data = {'role': self.association.role_set.get_benevole().id}
        response = client.post(url, data, follow=True)
        response = client.get('', follow=True)
        assert not count_text_in_content(response, "Modérations de bénévoles")

        # 5. Le bénévole est notifié
        assert len(mailoutbox) == 2
        assert foreign.email in mailoutbox[1].to
        assert foreign.get_short_name() in mailoutbox[1].body
        assert "Mise à jour de votre engagement" in mailoutbox[1].subject

        # 6. le bénévole a désormais accès
        client.logout()
        client.force_login(foreign)
        response = client.get(self.board_url)

        assert count_text_in_content(response, "Mes engagements associatifs")
        assert count_text_in_content(response, "Enregistrer une association")

        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert len(table_row) == 1
        assert "AssoTest" in str(table_row[0])
        assert "disabled" not in str(table_row[0])


@pytest.mark.django_db
class TestBoard:
    url = reverse_lazy('base:board')

    def test_get_anonymous(self, client):
        response = client.get(self.url, follow=True)
        assert response.status_code == 200
        assert is_login_view(response)

    def test_get_user_not_engaged(self, client, user):
        client.force_login(user)

        # the user is not engaged by those associations
        Association.objects.create(name='bling', conceal=True)
        Association.objects.create(name='blang')

        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, "Aucune donnée disponible")
        assert count_text_in_content(response, "Mes engagements associatifs")
        assert count_text_in_content(response, "Enregistrer une association")

    def test_get_user_engaged(self, client, association, engagement, user):
        client.force_login(user)

        response = client.get(self.url)
        assert response.status_code == 200
        assert count_text_in_content(response, "AssoTest")

    @pytest.mark.parametrize(
        'use, expected', [(None, 404), ('user', 200), ('staff', 404)]
    )
    def test_create_associated_self_benevalo_access(
        self, client, association, engagement, user, superuser, use, expected
    ):
        if use == 'staff':
            client.force_login(superuser)
        elif use:
            client.force_login(user)

        response = client.get(
            reverse(
                'base:association:self-benevalo:create', args=[association.id]
            )
        )
        assert response.status_code == expected

    def test_create_update_delete_associated_self_benevalo(
        self, client, association, category, engagement, user, superuser
    ):
        client.force_login(user)

        # 0. On fetch les engagements
        response = client.get(self.url)
        assert response.status_code == 200

        # 1. on créé
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a', class_='create_benevalo-link')[-1]
            .attrs['href']
        )
        response = client.get(url, follow=True)
        assert not count_text_in_content(response, "level")
        assert not count_text_in_content(response, "project")

        data = {
            'user': user.id,
            'title': 'bénévalo de test',
            'duration_1': '33',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }
        response = client.post(url, data, follow=True)
        assert response.status_code == 200

        association.benevalo_set.get(user=user, title='bénévalo de test')

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été ajouté" in messages[0]

        # 2. On a la liste

        table = str(bs(response.content, 'html.parser').find('tbody'))

        assert "33 minutes" in table
        assert "bénévalo de test" in table

        # 3. On update
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a')[-1]
            .attrs['href']
        )
        data = {
            'user': user.id,
            'title': 'bénévalo de lutte',
            'duration_1': '42',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }
        response = client.post(url, data, follow=True)

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été modifié" in messages[0]

        table = str(bs(response.content, 'html.parser').find('tbody'))
        assert "33 minutes" not in table
        assert "bénévalo de test" not in table
        assert "42 minutes" in table
        assert "bénévalo de lutte" in table

        # 4. on vérifie la date dans le formulaire  (BUG #20)
        response = client.get(url, follow=True)
        execution_date = (
            bs(response.content, 'html.parser')
            .find(id='id_execution_date')
            .attrs['value']
        )
        assert execution_date == date.today().strftime('%Y-%m-%d')

        # 5. On delete
        url = (
            bs(response.content, 'html.parser')
            .find('a', string=RE_DELETE_BTN)
            .attrs['href']
        )
        data = {'confirm': True}
        response = client.post(url, data, follow=True)
        assert response.status_code == 200
        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "» a été supprimé" in messages[0]

        table = str(bs(response.content, 'html.parser').find('tbody'))
        assert "Aucune donnée disponible." in table

    def test_create_associated_benevalo_with_one_level(
        self, client, association, level, category, engagement, user, superuser
    ):
        client.force_login(user)

        # 0. On fetch les engagements
        response = client.get(self.url)
        assert response.status_code == 200

        # 1. on créé
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a', class_='create_benevalo-link')[-1]
            .attrs['href']
        )
        response = client.get(url, follow=True)
        assert count_text_in_content(response, "level")
        assert not count_text_in_content(response, "project")

        data = {
            'user': user.id,
            'title': 'BenevaloTest',
            'duration_1': '33',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }

        response = client.post(url, data, follow=True)
        assert response.status_code == 200

        association.benevalo_set.get(user=user, title="BenevaloTest")

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été ajouté" in messages[0]

    def test_create_associated_benevalo_with_one_project(
        self,
        client,
        association,
        project,
        category,
        engagement,
        user,
        superuser,
    ):
        client.force_login(user)

        # 0. On fetch les engagements
        response = client.get(self.url)
        assert response.status_code == 200

        # 1. on créé
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a', class_='create_benevalo-link')[-1]
            .attrs['href']
        )
        response = client.get(url, follow=True)
        assert count_text_in_content(response, "project")
        assert not count_text_in_content(response, "level")

        data = {
            'user': user.id,
            'title': 'test',
            'duration_1': '33',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }

        response = client.post(url, data, follow=True)
        assert response.status_code == 200

        association.benevalo_set.get(user=user, title='test')

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été ajouté" in messages[0]

    def test_create_associated_benevalo_with_moderation(
        self, client, mailoutbox, association, category, manager, user, foreign
    ):
        association.moderate_benevalo = True
        association.save()

        engagement = association.engagement_set.create(
            user=foreign, role=association.role_set.get_benevole()
        )

        # 0. On fetch les engagements
        client.force_login(foreign)
        response = client.get(self.url)
        assert response.status_code == 200

        assert not count_text_in_content(response, "level")

        # 1. On fait la saisie
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a', class_='create_benevalo-link')[-1]
            .attrs['href']
        )
        data = {
            'user': user.id,
            'title': 'BenevaloTest',
            'duration_1': '33',
            'distance': '102',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }
        response = client.post(url, data, follow=True)
        assert response.status_code == 200

        benevalo = association.benevalo_set.get(
            user=foreign, title='BenevaloTest'
        )
        assert not benevalo.is_active

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "modération" in messages[0]

        # 2. Elle est visible mais suspendue en modération
        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert "(en modération)" in str(table_row[0])

        # 3. Elle est visible à un modérat⋅eur⋅rice
        client.logout()
        client.force_login(user)
        for urlname, args in [
            ('base:board', []),
            ('base:association:benevalo:index', [association.id]),
            (
                'base:association:engagement:benevalo',
                [association.id, engagement.id],
            ),
        ]:
            response = client.get(reverse(urlname, args=args))
            assert count_text_in_content(
                response, "Modérations de bénévolats en attente"
            )
            table_row = (
                bs(response.content, 'html.parser')
                .find('tbody')
                .find_all('tr')
            )
            assert len(table_row) == 1
            assert "foreign@example.org" in str(table_row[0])
            assert "BenevaloTest" in str(table_row[0])
            assert "Inspecter" in str(table_row[0])

        # Qui reçoit une notification
        assert len(mailoutbox) == 1
        assert user.email in mailoutbox[0].to
        assert user.get_short_name() in mailoutbox[0].body
        assert "Validation d'une nouvelle action" in mailoutbox[0].subject

        # 4. On donne suite
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a', class_='update-link')[-1]
            .attrs['href']
        )
        response = client.get(url, follow=True)

        # On prend tout et on post
        form = bs(response.content, 'html.parser').find("form")
        input_kv = [
            (i.get('name'), i.get('value'))
            for i in form.find_all('input')
            if i.get('value')
        ]
        select_kv = [
            (s.parent.get('name'), s.get('value'))
            for s in form.find_all('option', attrs={"selected": ""})
            if s.get('value')
        ]
        data = dict(input_kv + select_kv)

        response = client.post(url, data, follow=True)

        assert not count_text_in_content(
            response, "Modérations de bénévolats en attente"
        )

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été modifié" in messages[0]

        # 5. Le bénévole a désormais vue sur une saisie normale
        client.logout()
        client.force_login(foreign)
        response = client.get(
            reverse(
                'base:association:self-benevalo:index', args=[association.id]
            )
        )
        assert response.status_code == 200

        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert "(en modération)" not in str(table_row[0])
        assert "33 minutes" in str(table_row[0])
        assert "BenevaloTest" in str(table_row[0])


@pytest.mark.django_db
class TestUserList:
    @pytest.mark.parametrize(
        'use, expected',
        [(None, 404), ('user', 200), ('staff', 404), ('foreign', 404)],
    )
    def test_get_users(
        self,
        association,
        client,
        user,
        manager,
        superuser,
        foreign,
        use,
        expected,
    ):
        if use == 'staff':
            client.force_login(superuser)
        elif use == 'foreign':
            client.force_login(foreign)
        elif use:
            client.force_login(user)

        response = client.get(
            reverse('base:association:user', args=[association.id])
        )
        assert response.status_code == expected

        if response.status_code == 200:

            table_row = (
                bs(response.content, 'html.parser')
                .find('tbody')
                .find_all('tr')
            )
            assert len(table_row) == 1
            assert "Dirigeant" in str(table_row[-1])
            assert "user@example.org" in str(table_row[-1])


@pytest.mark.django_db
class TestBenevalo:
    @pytest.mark.parametrize(
        'use, expected',
        [(None, 404), ('user', 200), ('staff', 404), ('foreign', 404)],
    )
    def test_get_index(
        self,
        association,
        client,
        user,
        manager,
        superuser,
        foreign,
        use,
        expected,
    ):
        if use == 'staff':
            client.force_login(superuser)
        elif use == 'foreign':
            client.force_login(foreign)
        elif use:
            client.force_login(user)

        response = client.get(
            reverse('base:association:benevalo:index', args=[association.id])
        )
        assert response.status_code == expected

    def test_index_table_footer(
        self, client, association, category, user, engagement
    ):
        user.benevalo_set.create(
            title="test 1",
            duration=timedelta(0, 1260),
            distance=21,
            execution_date=date.today(),
            association=association,
            category=category,
        )

        user.benevalo_set.create(
            title="test 2",
            duration=timedelta(0, 1260),
            distance=21,
            execution_date=date.today() - timedelta(1),
            association=association,
            category=category,
        )

        client.force_login(user)
        response = client.get(
            reverse(
                'base:association:self-benevalo:index', args=[association.id]
            )
        )
        assert response.status_code == 200
        table_foot = bs(response.content, 'html.parser').find('tfoot')
        assert "42 minutes" in str(table_foot)
        assert "42,00 km" in str(table_foot)
        assert "2 éléments" in str(table_foot)
        assert "sur 1 jour" in str(table_foot)

    def test_index_table_export(
        self, client, association, project, level, category, user, engagement
    ):
        user.benevalo_set.create(
            title="test 1",
            duration=timedelta(0, 1260),
            distance=21,
            execution_date=date.today(),
            association=association,
            category=category,
        )

        user.benevalo_set.create(
            title="test 2",
            duration=timedelta(0, 1260),
            distance=21,
            execution_date=date.today() - timedelta(1),
            execution_end_date=date.today() + timedelta(1),
            association=association,
            category=category,
            level=level,
            project=project,
            description='mise sous plis des convocations',
            is_active=False,
        )

        client.force_login(user)
        response = client.get(
            reverse(
                'base:association:self-benevalo:index', args=[association.id]
            )
            + '?_export=json'
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                'Titre': "test 1",
                'Durée': "21 minutes",
                'Heures, en décimal': "0,35",
                'Distance': "21,00",
                'Date': date.today().isoformat(),
                'Association': str(association),
                'Bénévole': str(user),
                'Catégorie': str(category),
                'Niveau': None,
                'Description': None,
                'Projet': None,
                'Date de fin': None,
                'En attente de modération': False,
            },
            {
                'Titre': "test 2",
                'Durée': "21 minutes",
                'Heures, en décimal': "0,35",
                'Distance': "21,00",
                'Date': (date.today() - timedelta(1)).isoformat(),
                'Association': str(association),
                'Bénévole': str(user),
                'Catégorie': str(category),
                'Niveau': str(level),
                'Description': 'mise sous plis des convocations',
                'Projet': str(project),
                'Date de fin': (date.today() + timedelta(1)).isoformat(),
                'En attente de modération': True,
            },
        ]

    def test_post_create_update_delete(
        self, client, association, category, manager, user, foreign
    ):
        client.force_login(user)

        # 1. on fetch les benevalos de l'asso
        url = reverse('base:association:benevalo:index', args=[association.id])
        response = client.get(url)
        assert response.status_code == 200
        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert len(table_row) == 1
        assert "Aucune donnée disponible." in str(table_row[-1])

        # 2. on créé un benevalo dans l'asso
        url = (
            bs(response.content, 'html.parser')
            .find(class_='title-buttons')
            .find('a')
            .attrs['href']
        )
        data = {
            'user': user.id,
            'title': 'benevalo de test',
            'duration_1': '42',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }
        response = client.post(url, data, follow=True)
        assert response.status_code == 200
        assert count_text_in_content(response, "benevalo de test")
        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "« benevalo de test » a été ajouté" in messages[0]

        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert len(table_row) == 1
        assert "benevalo de test" in str(table_row[-1])

        # 3. on tente l'update
        url = (
            bs(response.content, 'html.parser')
            .find('tbody')
            .find_all('a')[-1]
            .attrs['href']
        )
        data = {
            'user': foreign.id,
            'title': 'benevalo de lutte',
            'duration_1': '42',
            'execution_date': date.today(),
            'association': association.id,
            'category': category.id,
        }
        response = client.post(url, data, follow=True)
        assert response.status_code == 200

        assert count_text_in_content(
            response, "Ce choix ne fait pas partie de ceux disponibles."
        )
        execution_date = (
            bs(response.content, 'html.parser')
            .find(id='id_execution_date')
            .attrs['value']
        )
        assert execution_date == date.today().strftime('%Y-%m-%d')

        # 4. Bien reçu, on retente en intégrant le foreign user
        association.engagement_set.create(
            user=foreign, role=association.role_set.get_benevole()
        )
        response = client.post(url, data, follow=True)

        assert not count_text_in_content(response, "benevalo de test")
        assert count_text_in_content(response, "benevalo de lutte")
        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été modifié" in messages[0]

        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert len(table_row) == 1
        assert "benevalo de lutte" in str(table_row[-1])

        # 5. on delete
        response = client.get(url, follow=True)
        url = (
            bs(response.content, 'html.parser')
            .find('a', string=RE_DELETE_BTN)
            .attrs['href']
        )
        data = {'confirm': True}
        response = client.post(url, data, follow=True)
        assert response.status_code == 200

        messages = [m.message for m in get_messages(response.wsgi_request)]
        assert len(messages) == 1
        assert "a été supprimé" in messages[0]

        table_row = (
            bs(response.content, 'html.parser').find('tbody').find_all('tr')
        )
        assert len(table_row) == 1
        assert "Aucune donnée disponible." in str(table_row[-1])


@pytest.mark.django_db
class TestEngagedBenevalo:
    @pytest.mark.parametrize(
        'use, expected',
        [(None, 404), ('user', 200), ('staff', 404), ('foreign', 404)],
    )
    def test_get_index(
        self,
        association,
        client,
        user,
        manager,
        superuser,
        foreign,
        use,
        expected,
    ):
        if use == 'staff':
            client.force_login(superuser)
        elif use == 'foreign':
            client.force_login(foreign)
        elif use:
            client.force_login(user)

        response = client.get(
            reverse(
                'base:association:engagement:benevalo',
                args=[association.id, association.get_engagement(user).id],
            )
        )
        assert response.status_code == expected
