from django.urls import path, resolve, reverse, set_urlconf
from django.urls.exceptions import NoReverseMatch

import pytest

from ..urls import crud_include
from . import views


@pytest.fixture
def crud_urlconf():
    def set_crud_urlconf(class_prefix, *args, **kwargs):
        views_module = kwargs.pop('views_module', views)
        set_urlconf(
            (
                path(
                    '',
                    crud_include(views_module, class_prefix, *args, **kwargs),
                ),
            )
        )

    yield set_crud_urlconf

    set_urlconf(None)


class TestCrudInclude:
    def test_reverse_names(self, crud_urlconf):
        crud_urlconf('MyModel', 'my-model')
        assert reverse('my-model:index') == '/'
        assert reverse('my-model:create') == '/create/'
        assert reverse('my-model:detail', args=(1,)) == '/1/'
        assert reverse('my-model:update', args=(1,)) == '/1/update/'
        assert reverse('my-model:delete', args=(1,)) == '/1/delete/'

    def test_resolve_classes(self, crud_urlconf):
        crud_urlconf('MyModel', 'my-model')
        assert resolve('/').func.view_class == views.MyModelIndex
        assert resolve('/create/').func.view_class == views.MyModelCreate
        assert resolve('/1/').func.view_class == views.MyModelDetail
        assert resolve('/1/update/').func.view_class == views.MyModelUpdate
        assert resolve('/1/delete/').func.view_class == views.MyModelDelete

    def test_views_module_name(self, crud_urlconf):
        crud_urlconf('MyModel', '', views_module='{}.views'.format(__package__))
        assert reverse('index') == '/'

    def test_empty_namespace(self, crud_urlconf):
        crud_urlconf('MyModel', '')
        assert reverse('index') == '/'

    def test_custom_classes_map(self, crud_urlconf):
        crud_urlconf(
            '', classes_map=[('foo/', 'foo', 'Foo'), ('bar/', 'bar', 'Bar')]
        )
        assert reverse('foo') == '/foo/'
        # ... but Bar doesn't exist
        with pytest.raises(NoReverseMatch):
            reverse('bar')

    def test_no_views_found(self, crud_urlconf):
        with pytest.raises(ValueError):
            crud_urlconf('Empty', '')
